#include <fstream>
using std::ofstream;
#include <iostream>
using std::cout;
using std::endl;
#include <math.h>

#include "protein.H"
#include "roadmapConstruction.H"

///@name MCSimulation.C
//@{

typedef enum foldingResult{folded, unfolded, notFoldedNeitherUnfolded};

//
bool metropolisAccept (float startEnergy, float endEnergy)
{
  if (endEnergy < startEnergy)
    return true;
  else
    {
      float rN = random() / MY_RAND_MAX;
      if (rN < exp (- (endEnergy - startEnergy) * inv_kT))
	{
	  return true;
	}
    }
  return false;
}


/**
   finds a config other than (around) the original, returns it; unless
   numSteps exceeds maxNumStepsBeforeExitingThisFunction, in which case
   returns NULL.

   updates numSteps, where a step is defined as a move trial in MC simulation (which may be a stay in the current node.)

   maxNumStepsInSamplingAConfig : tries at most this many trials in the "sampling Around" of the original protein.

   "sampling around" means that the new config will be within +-deltaX,deltaY
   of the original.

*/

Protein *  MC_OneJump (const Protein * startP, 
		       int           * numSteps,
		       int maxNumStepsInSamplingAConfig         = (int) 1E3,
		       int maxNumStepsBeforeExitingThisFunction = (int) 1E6)

{
  Protein * retval = NULL;
  float newEnergy, startEnergy = startP->getEnergy();
  bool      metropolisCrit;

  assert (startEnergy < Protein::maxEnergy);
  assert ((*numSteps) < maxNumStepsBeforeExitingThisFunction);
  do 
    {
      if (retval != NULL)
	delete retval;

      retval    = new Protein(-1, startP, numSteps, 
			      maxNumStepsInSamplingAConfig);
      newEnergy = retval->getEnergy();
      
      assert (newEnergy < Protein::maxEnergy);
      
      metropolisCrit = metropolisAccept(startEnergy, newEnergy);

    } 
  while ((!metropolisCrit) && 
	 (*numSteps < maxNumStepsBeforeExitingThisFunction));
  
  if (metropolisCrit)
    return retval;
  else
    {
      assert (retval != NULL);
      delete retval;
      return NULL;
    }
}



/*
  sets rmsd2Folded, rmsd2Unfolded; foundFoldedState and foundUnfoldedState.
  first call should have  minDistFolded, minDistUnfolded, foldedProtein and
  unfoldedProtein set to desired values and firstCall set to true. 
  Later calls should not set these parameters.
*/

void checkLocation(Protein * currentProtein, float *dist2Folded, 
		   float *dist2Unfolded, bool *foundFoldedState, 
		   bool * foundUnfoldedState,
		   bool setVariables = false,
		   float minDistFolded=0.0, float minDistUnfolded=0.0,
		   Protein * foldedProtein= NULL, 
		   Protein * unfoldedProtein= NULL)
{
  static float f_minDistFolded = -1.0, f_minDistUnfolded = 0.0;
  static Protein * p_foldedP = NULL, * p_unfoldedP=NULL;
  if (setVariables)
    {
      assert (f_minDistFolded == -1.0);
      assert (minDistFolded != 0.0);
      assert (minDistUnfolded != 0.0);
      assert (foldedProtein != NULL);
      assert (unfoldedProtein != NULL);
      
      f_minDistFolded   = minDistFolded;
      f_minDistUnfolded = minDistUnfolded;
      p_foldedP         = foldedProtein;
      p_unfoldedP       = unfoldedProtein;
    }
  
  assert (f_minDistFolded > 0);
  assert (f_minDistUnfolded > 0);
  assert (p_foldedP != NULL);
  assert (p_unfoldedP != NULL);

  *dist2Folded   = p_foldedP  ->computeDistance(currentProtein);
  *dist2Unfolded = p_unfoldedP->computeDistance(currentProtein);

  if (*dist2Folded < f_minDistFolded)
    {
      *foundFoldedState = true;
    }
  else
    assert (*foundFoldedState == false);
  
  if (*dist2Unfolded < f_minDistUnfolded)
    {
      *foundUnfoldedState = true;
    }
  else
    assert (*foundUnfoldedState == false);
}

//creates the file, prints basic info there.
void  preparePfoldOutputFile(int startIndex, ofstream * fout, 
			     float minRMSDFolded, float minRMSDUnfolded, 
			     int maxNumMCSimulations, int maxIter)
{
  char filename[1000];
  
  sprintf(filename, "pfold%d.txt", startIndex);
  cout << "check " << filename << endl;

  fout->open(filename);
  *fout << "%pfold computation, " << minRMSDFolded << " and " << minRMSDUnfolded << " radii, " << maxNumMCSimulations << " sim./node, starting from node # " << startIndex << endl;
  *fout << "%max " << maxIter << " iterations / node." << endl;
  *fout << "%format: nodeIndex #folded #unfolded #numMCsimulations" << endl;

}

foldingResult findMCSimulationReturnValue(bool foundFoldedState, 
					  bool foundUnfoldedState)
{
  foldingResult retval;
  
  if (((foundFoldedState) && (foundUnfoldedState)) || 
      ((!foundFoldedState) && (!foundUnfoldedState)))
    {
      retval = notFoldedNeitherUnfolded;
      //cout << "not folded neither unfolded." << endl;
    }
  else if (foundUnfoldedState)
    {
      retval = unfolded;
      //cout << "unfolded." << endl;
    }
  else if (foundFoldedState)
    {
      retval = folded;
      //cout << "folded." << endl;
    }

  return retval;
}

/**
   does MC from startNodeIndex. stops when it gets close to either
   foldedStateIndex or unfoldedStateIndex. Does a max. of maxIter
   iterations.

   returns foldingResult.
*/	  
foldingResult MCSimulate (Protein * startP,
			  const int maxIter,
			  Protein * foldedP, Protein * unfoldedP,
			  float minDistFolded, float minDistUnfolded)
{
  Protein *     p2, *p1;
  float         rmsd2Folded, rmsd2Unfolded;
  bool          foundFoldedState     = false, foundUnfoldedState = false;
  int           numIter = 0, maxNumStepsInSamplingAConfig = 1000;
  foldingResult retval;

  p1       = startP;
  p2       = NULL;
  
  // note that this MC simulation is assuming that checkLocation is called
  // prior to the execution if this function to set that functions static 
  // variables, so it is not set again. For using this function as stand 
  // alone one should use the extra parameters as set in pfoldComputation 
  // routine.

  checkLocation (p1, &rmsd2Folded, &rmsd2Unfolded, 
		 &foundFoldedState, 
		 &foundUnfoldedState);
  while ((numIter < maxIter) && (!foundFoldedState) && (!foundUnfoldedState))
    {
      p2 = MC_OneJump(p1, &numIter, maxNumStepsInSamplingAConfig, maxIter);
      if (p2 == NULL)
	{
	  assert (numIter >= maxIter);
	}
      else
	{
	  checkLocation (p2, &rmsd2Folded, &rmsd2Unfolded, 
			 &foundFoldedState, &foundUnfoldedState);
	}
      if (p1 != startP) // don't want to delete a node of the roadmap, although nothing would happen unless i want to use that node again
	{
	  assert (p1 != NULL);
	  delete p1;
	}
      p1 = p2;
    }

  if ((p1 != NULL) && (p1 != startP))
    {
      delete p1;
    }
  
  
  // finding return value...
  retval = findMCSimulationReturnValue(foundFoldedState, foundUnfoldedState);

  //cout << "did " << numIter << " iterations." << endl;
  return retval;
}


/**
   computes pfold for node # startIndex.
*/
void MCpfoldComputation (Protein * foldedP, Protein * unfoldedP,
		       Protein * startP,
		       float minDistFolded,
		       float minDistUnfolded,
		       int maxIter,
		       int maxNumMCSimulations,
		       int startIndex)
  
{
  int             numFolded = 0, numUnfolded = 0, numMCSimulations = 0;
  foldingResult   simulationResult;
  ofstream        fout;
  bool            callingCheckLocationFirstTime    = true, dummyBool = false;
  float           dummyFloat;
  
  assert (maxIter > 0);
  assert (maxNumMCSimulations > 0);
  assert (maxNumMCSimulations % 100 == 0);//arbitrary condition can be changed or removed

  preparePfoldOutputFile(startIndex, &fout, minDistFolded, 
			 minDistUnfolded, maxNumMCSimulations, maxIter);

  cout << "starting from node # " << startIndex << endl;
  
  checkLocation (startP, &dummyFloat, &dummyFloat,
		 &dummyBool, &dummyBool,
		 callingCheckLocationFirstTime,
		 minDistFolded, minDistUnfolded,
		 foldedP,  unfoldedP);


  while (numMCSimulations < maxNumMCSimulations)
    {
      simulationResult = MCSimulate(startP, maxIter, 
				    foldedP,
				    unfoldedP,
				    minDistFolded, minDistUnfolded);
      if (simulationResult == folded)
	{
	  numFolded ++;
	}
      else if (simulationResult == unfolded)
	{
	  numUnfolded ++;
	}
      else
	{
	  //cout << " MC simulation did not fold or unfold." << endl;
	}
      numMCSimulations++;
      if (numMCSimulations % 100 == 0)
	{
	  fout << startIndex << " " << numFolded << " " << numUnfolded << " ";
	  fout << numMCSimulations  << endl;
	}
    }

  cout << "node # " << startIndex << " pfold after " << numMCSimulations << " MC simulations is : " << (numFolded * 1.0/ (numFolded + numUnfolded)) << endl;
  cout << " did not fold nor unfold " << numMCSimulations - numFolded - numUnfolded << " times." << endl;
  fout.close();

}


//@}
