#include <vector>
#include <iostream>
using std::cout;
using std::endl;
using std::cerr;

#include "tree.H"
#include "analysis.H"
#include "roadmapConstruction.H"
#include "protein.H"

/**@name tree.C*/
//@{

//--------------------treeclass methods ----------------------

treeClass::treeClass() {

  root     = NULL;
  nextTree = NULL;
  prevTree = NULL;
  numNodesInTree = 0;

}

///
treeClass::treeClass(nodeClass * rootNode, int numNodes, 
		     ccInfoType * ccInfo)
{
  this->root           = rootNode;
  this->numNodesInTree = numNodes;
  this->prevTree       = ccInfo->globalLastTree;
  this->myTreeName     = ccInfo->globalTreeName;
  this->nextTree       = NULL;

  ccInfo->globalTreeName++;
  ccInfo->numTrees++;
  rootNode->myConfig->myTree = this;
  appendNewTreePtr(ccInfo, this);
  
}


///
void treeClass::printTrees(int numNodes, Protein ** allNodes) 
{
  cout << "Tree name: " << myTreeName << " Tree Size: " << numNodesInTree << endl;
  initNodeColor(numNodes, 0, allNodes);

  root->printTree();

  if (nextTree != NULL) 
    {
      nextTree->printTrees(numNodes, allNodes);
    }
}

///returns the indices of nodes in current tree.
void treeClass::findElementsInTree(set<int> & elements, int numNodes, 
				   Protein ** allNodes)
{
  elements.clear();
  assert (elements.empty());
  initNodeColor(numNodes, 0, allNodes);
  root->printTree();
  for (int i = 0; i < numNodes; i++)
    {
      if (allNodes[i]->myNode->nodeColor == 1)
	{
	  elements.insert(i);
	}
    }
}

/**
   moves the nodes in the (this) connected component 
   to replacementTree connected component. updates 
   ccInfo.
   
   replacementTree can be NULL, in which case 
   the nodes' myTree pointer is set to 
   NULL.

   the caller after calling should delete the tree with a delete command.
 */
void treeClass::removeTree (treeClass* replacementTree,
			    ccInfoType * ccInfo,
			    Protein   ** allNodes, 
			    const int    numNodes)
{

  if (replacementTree != NULL)
    replacementTree->numNodesInTree += numNodesInTree;

  initNodeColor( numNodes,0 , allNodes);
  
  root->replaceTreePtr(replacementTree, NULL);
  
  if (prevTree != NULL) 
    {
      prevTree->nextTree = nextTree;
    }
  else 
    {
      ccInfo->globalFirstTree = nextTree;
    }

  if (nextTree != NULL) 
    {
      nextTree->prevTree = prevTree;
    }
  else 
    {
      ccInfo->globalLastTree = prevTree;
    }
  
  ccInfo->numTrees--;
  
}

//------------------ptrList methods --------------------------

ptrList::ptrList() {

  nextPtr = prevPtr = NULL;
  node = NULL;
  pathWeight = 0.0;
  
}

//------------------nodeClass methods-------------------------

nodeClass::nodeClass() {

  firstLink = NULL;
  lastLink = NULL;
  myConfig = NULL;
  pathPtr = NULL;
  nodeColor = (char)(0);

}


///adds a link, increments the owning protein's numLinks by 1.
void nodeClass::addNodeLink(nodeClass *thelink, float wt) 
{

  ptrList *tempPtrList;

  tempPtrList             = new ptrList;

  tempPtrList->node       = thelink;
  tempPtrList->pathWeight = wt;

  if (lastLink == NULL) {
    
    lastLink = tempPtrList;
    firstLink = tempPtrList;
  }
  else {
    tempPtrList->prevPtr = lastLink;
    lastLink   ->nextPtr = tempPtrList;
    lastLink             = tempPtrList;
  }
  
  myConfig->setNumLinks(myConfig->getNumLinks() + 1);

}

///removes all edges going out of this node, resets its edge parameters accordingly.
void nodeClass::removeAllLinks()
{
  ptrList * tempEdge, * edge = firstLink;
  
  while (edge != NULL)
    {
      tempEdge = edge;
      edge     = edge->nextPtr;
      delete tempEdge;
    }
  firstLink = lastLink = NULL;
  pathPtr   = NULL;  // this may not be necessary.
  nodeColor = (char)(0); 
  myConfig->setNumLinks(0);
}

//to replace all nodes tree pointer, connected to this node. To be used by treeClass::removeTree.
void nodeClass::replaceTreePtr(treeClass *replacementTree, 
			       nodeClass *fromNode) 
{
  ptrList *linkPtr;

  myConfig->myTree = replacementTree;
  nodeColor        = (char)(1);
  linkPtr          = firstLink;
  
  while (linkPtr != NULL) {
    // the first part of the condition is not necessary, i think,
    // since when part 1 is false, part2 is false too.
    if ((linkPtr->node != fromNode) && 
	(linkPtr->node->nodeColor == (char)(0)))
      {
	linkPtr->node->replaceTreePtr(replacementTree, this);
      }
    linkPtr = linkPtr->nextPtr;
  }
}

/// nodeColor of visited nodes becomes 1. 
void nodeClass::markAllNodesInTheTree(int * numNodesInTree)
{
  ptrList *linkPtr = firstLink;
  nodeColor        = (char)(1);
  
  (*numNodesInTree) ++;
  while (linkPtr != NULL) {
    if 	(linkPtr->node->nodeColor == (char)(0))
      {
	linkPtr->node->markAllNodesInTheTree(numNodesInTree);
      }
    linkPtr = linkPtr->nextPtr;
  }

}

///prints the index of every node in the same tree as (this)
void nodeClass::printTree()
{
  ptrList *linkPtr = firstLink;
  nodeColor        = (char)(1);

  //cout << "node index = " << myConfig -> getIndex() << endl;

  while (linkPtr != NULL) {
    if 	(linkPtr->node->nodeColor == (char)(0))
      {
	linkPtr->node->printTree();
      }
    linkPtr = linkPtr->nextPtr;
  }
  
}

///removes link from (this) to (toNode). returns true if such a connection exists (and is removed), false o.w. does not do anything with the trees.
bool nodeClass::removeConnectionTo(nodeClass * toNode)
{
  ptrList * tempPtrList = firstLink;
  bool retval;
  
  while ((tempPtrList != NULL) && (tempPtrList -> node != toNode) )
    {
      tempPtrList = tempPtrList ->nextPtr;
    }

  if (tempPtrList == NULL)
    retval = false;
  else
    {
      retval = true;
      assert (tempPtrList != NULL);
      
      if ((tempPtrList == lastLink) || (tempPtrList == firstLink))
	{
	  if (tempPtrList == lastLink)
	    {
	      lastLink = tempPtrList ->prevPtr;
	      if (lastLink != NULL)
		lastLink->nextPtr = NULL;
	    }
	  if (tempPtrList == firstLink)
	    {
	      firstLink = firstLink->nextPtr;
	      if (firstLink != NULL)
		firstLink->prevPtr = NULL;
	    }
	}
      else 
	{
	  // tempPtrList is not firstLink and not lastLink. somewhere in the middle
	  assert (tempPtrList->prevPtr != NULL);
	  tempPtrList->prevPtr->nextPtr = tempPtrList->nextPtr;

	  assert (tempPtrList->nextPtr != NULL);
	  tempPtrList->nextPtr->prevPtr = tempPtrList->prevPtr;
	}
      delete tempPtrList;
    }
  return retval;
}



//------------------other helpers ------------------------------

//used by treeClass constructor
void  appendNewTreePtr(ccInfoType * ccInfo, treeClass * newTreePtr)
{
  if (ccInfo->globalLastTree == NULL) 
    {
      ccInfo->globalLastTree  = newTreePtr;
      ccInfo->globalFirstTree = newTreePtr;
    }
  else 
    {
      ccInfo->globalLastTree->nextTree = newTreePtr;
      ccInfo->globalLastTree           = newTreePtr;
    }
}


/**
  returns true if n1 is connected to n2.
*/
bool connected (int n1, int n2, Protein ** allNodes)
{
  ptrList * linkPtr = allNodes[n1]->myNode->firstLink;
  
  while (linkPtr != NULL)
    {
      if (linkPtr->node->myConfig->getIndex() == n2)
	{
	  //	  cout << n1 << " is already connected to : " << n2 << endl;
	  return true;
	}
      linkPtr = linkPtr->nextPtr;
    }
  return false;
}


/** 
    returns neighbors of node # startNode.
*/

void findNeighbors (vector<int> &neighborIndices, int startNode, 
		    Protein ** allNodes) 
{
  vector<float> dummyFloatVector;

  findNeighborsAndTransitionProbabilities(neighborIndices, dummyFloatVector,
					  startNode, allNodes);
}


///returns num neighbors of node # i.
unsigned int findNumNeighbors (int i, Protein ** allNodes)
{

  vector<float> dummyFloatVector;
  vector<int>   dummyIntVector;

  findNeighborsAndTransitionProbabilities(dummyIntVector, dummyFloatVector,
					  i, allNodes);

  assert (dummyIntVector.size() == dummyFloatVector.size());
  return dummyIntVector.size();
}


/**
  finds and returns both node indices that the input node is connected
  to, as well as the edge weights.
*/
void findNeighborsAndTransitionProbabilities (vector<int>   &neighborIndices,
					      vector<float> &tranProbs,
					      int startNode,
					      Protein ** allNodes)
{
  ptrList * linkPtr;
  
  linkPtr  = allNodes[startNode]->myNode->firstLink;

  neighborIndices.resize(0);
  tranProbs    .resize(0);

  while (linkPtr != NULL) 
    {
      neighborIndices.push_back (linkPtr->node->myConfig->getIndex());
      tranProbs      .push_back (linkPtr->edgeTransitionProbability);
      linkPtr       = linkPtr->nextPtr;
    }
}


// to be used by makeLinkBetweenNodes. returns the change in number of cc's.
int updateTreesOfConnectedPairOfNodes(int n1, int n2, int numNodes, 
				       Protein ** allNodes, 
				       ccInfoType * ccInfo)
{
  int retval = 0;
  treeClass * tempTreePtr;
  
  if ((allNodes[n2]->myTree == NULL) && (allNodes[n1]->myTree == NULL)) 
    {
      retval = 1;
     
      tempTreePtr = new treeClass(allNodes[n2]->myNode,
				  2,
				  ccInfo);

      allNodes[n1]->myTree = tempTreePtr;
      
    }
  
  else if ((allNodes[n2]->myTree == NULL) && (allNodes[n1]->myTree != NULL)) 
    {
      allNodes[n2]->myTree = allNodes[n1]->myTree;
      allNodes[n1]->myTree->numNodesInTree++;
      assert (retval == 0);
      //retval stays 0
    }
  
  else if ((allNodes[n2]->myTree != NULL) && (allNodes[n1]->myTree == NULL)) 
    {
      // first updating cog.
      allNodes[n1]->myTree = allNodes[n2]->myTree;
      allNodes[n2]->myTree->numNodesInTree++;
      assert (retval == 0);
    }
  else if 
    ((allNodes[n2]->myTree != NULL) && 
     (allNodes[n1]->myTree != NULL) && 
     (allNodes[n2]->myTree != allNodes[n1]->myTree)) 
    {
      tempTreePtr = allNodes[n2]->myTree;
      tempTreePtr->removeTree(allNodes[n1]->myTree, 
			      ccInfo,
			      allNodes, numNodes);
      delete tempTreePtr;
      retval = -1;
    }
  return retval;
}

/*
   The link between n1 and n2 is removed. Checks to see whether 
   there is a new connected component because of this removal and 
   creates it appropriately.

   to be used by removeLinkBetweenAndUpdateTrees.
*/
void updateTreesOfDisconnectedPairOfNodes (int n1, int n2, int numNodes, 
					   Protein ** allNodes, 
					   ccInfoType * ccInfo)
{
  treeClass * newTreePtr, * currentTreePtr = allNodes[n1]->myTree;
  int unreachableNode, numNodesRemainingInOrigTree = 0;

  initNodeColor(numNodes, 0, allNodes);
  currentTreePtr->root->markAllNodesInTheTree(&numNodesRemainingInOrigTree);

  if (numNodesRemainingInOrigTree == currentTreePtr->numNodesInTree)
    {
      //cout << " nothing to be done. the disconnection of the edge from n1 to n2 did not create a new cc." << endl;
      return;
    }

  currentTreePtr->numNodesInTree = numNodesRemainingInOrigTree;
  
  // we are sure that there is another cc arising from the removal of the edge from n1 to n2.
  // now we will create a tree for this cc and set the myTree ptr of all the nodes in that cc to that tree.
  
  if (allNodes[n1]->myNode->nodeColor == 1)
    {
      unreachableNode = n2;
    }
  else
    unreachableNode   = n1;

  assert (allNodes[unreachableNode]->myNode->nodeColor == 0);
  assert (allNodes[unreachableNode]->myTree    = currentTreePtr);
  
  initNodeColor(numNodes, 0, allNodes);
  int numNodesInNewTree = 0;
  allNodes[unreachableNode]->myNode->markAllNodesInTheTree(&numNodesInNewTree);
  // now all nodes in the new cc are "colored".
  newTreePtr = new treeClass(allNodes[unreachableNode]->myNode,
			  numNodesInNewTree,
			  ccInfo);
			  
  int count = 0; // for debugging, counting num nodes in the new cc.
  for (int i = 0; i < numNodes; i++)
    {
      if (allNodes[i]->myNode->nodeColor == 1)
	{
	  allNodes[i]->myTree = newTreePtr;
	  count++;
	}
    }
  cout << "there are " << count << " nodes in the new tree." << endl;
  assert (count == numNodesInNewTree);

}

/// removes link between node n1 and n2. Then updates the tree structure. returns true if such a connection is removed, false o.w. (i.e. no such connection existed in the first place)
bool removeLinkBetweenAndUpdateTrees (int n1, int n2, int numNodes, 
				      Protein ** allNodes, ccInfoType * ccInfo)
{
  bool remove1, remove2;
  
  //cout << "removing link between " << n1 << "  and " << n2 << endl;

  if (allNodes[n1]->myTree != allNodes[n2]->myTree)
    return false;
  

  remove1 = allNodes[n1]->myNode->removeConnectionTo(allNodes[n2]->myNode);
  remove2 = allNodes[n2]->myNode->removeConnectionTo(allNodes[n1]->myNode);

  assert (remove1 == remove2);
  
  if (remove1)
    updateTreesOfDisconnectedPairOfNodes (n1, n2, numNodes, allNodes, ccInfo);

  return remove1;
}

///returns the pointer to the largest connected component, if there are multiple such connected components returns NULL.
treeClass * findBiggestCC(const ccInfoType * ccInfo)
{
  treeClass *tempTreePtr, *retval;
  int numNodesInTree, maxNumNodes = 0;
  bool moreThanOneBiggestCC = false; // true if there are two connected components of equal and largest number of nodes

  tempTreePtr = ccInfo->globalFirstTree;

  while (tempTreePtr != NULL)
    {
      numNodesInTree =  tempTreePtr->numNodesInTree;
      if (numNodesInTree >= maxNumNodes)
	{
	  if (numNodesInTree > maxNumNodes)
	    {
	      maxNumNodes          = numNodesInTree;
	      moreThanOneBiggestCC = false;
	      retval               = tempTreePtr;
	    }
	  else
	    {
	      assert (numNodesInTree==maxNumNodes);
	      moreThanOneBiggestCC = true;
	    }
	}
      tempTreePtr = tempTreePtr->nextTree;
    }
  if (moreThanOneBiggestCC) 
    {
      cerr << "there are more than one connected component with largest size (of size : " << maxNumNodes << " ). Thus returning NULL." << endl;
      retval = NULL;
    }
  return retval;
}
//@}
