#include <fstream>
#include <math.h>
#include <stdio.h>
#include <assert.h>
#include <vector>
#include <iostream>
using std::cout;
using std::endl;
#include "tree.H"
#include "protein.H"


/**@name protein.C */
//@{

vector<float>     Protein::alphas, Protein::taus, Protein::xos, Protein::yos;
// used in the energy computation
const float       Protein::maxEnergy   =   6.0; 
const float       Protein::minEnergy   =  -6.0; 
const float Protein::maxX       = 100.0; // all x between -maxX..maxX.
const float Protein::maxY       = 100.0; // all y between -maxY..maxY.
const float Protein::resolution = 1.0;   // distance between intermed configs
const float Protein::deltaX     = 2.0;   // when a node is sampled around a 
                                         //seed node, the new node has x
                                         //within deltaX of the original.
const float Protein::deltaY     = 2.0; 
const unsigned int   Protein::NUMDOF      = 2;     // num. DOF of the system.
int   Protein::numEnergyComputations      = 0;


//---------------------------------------------------------------------------
//--------constructors, destructors------------------------------------------

/** 
    samples a new config, and guarantees it is not of energy > maxEnergy.
*/

Protein::Protein(int inputIndex)
{
  nodeClass * newNode;
  myTree   = NULL;
  myNode   = NULL;
  numLinks = 0;
  index    = inputIndex;
  
  vector<float> coords = generateRandomConfig();
  x = coords[0]; y = coords[1];
  computeEnergy();
  assert (energy < maxEnergy);
  coords2Params(params, coords);

  newNode                 = new nodeClass;
  newNode->myConfig       = this;
  this   ->setMyNode(newNode);
}


Protein::~Protein()
{
  if (myNode != NULL)
    delete myNode;
}


/** 
    samples around original node. Returned node is guaranteed to be within 
    bounds of the landscape and collision free.
    also does the node creation.

    the new node coords (x,y) are within protein::delta(x,y) of the original.
    index : index of protein in allNodes array.

    if it can't find a config of low energy within maxIter trials,
    it aborts.

    numIterOverall is also updated.
	
*/
Protein::Protein(int inputIndex, const Protein * original, 
		 int * numIterOverall,
		 const int maxIter )
{
  nodeClass *   newNode;
  vector<float> origCoords(2), newCoords;
  int numIterInThisFunction = 0;

  assert (maxIter > 0);

  origCoords[0] = original->getX(); origCoords[1] = original->getY(); 
  do
    {
      sampleAroundCoords            (newCoords, origCoords);
      coords2Params                 (params, newCoords);

      assert (newCoords.size() == 2);
      x = newCoords[0]; y = newCoords[1];
      
      computeEnergy                 ();
      if (numIterOverall != NULL)
	(*numIterOverall)++;
      numIterInThisFunction++;
    }
  while ((energy >= maxEnergy) && (numIterInThisFunction < maxIter));
	
  if (energy >= maxEnergy)
    {
      cout << "have done " << numIterInThisFunction << " iterations but couldn't generate a collision-free configuration around original." << endl;
      cout << "original conformation is : " <<  *original << endl;
      exit(0);
    }
  
  myTree   = NULL;
  myNode   = NULL;
  numLinks = 0;
  index    = inputIndex;
  newNode                 = new nodeClass;
  newNode->myConfig       = this;
  setMyNode(newNode);
}



/**
   creates a node located at coords. The energy should be < maxEnergy, o.w. it aborts.
*/

Protein::Protein (int inputIndex, const vector<float>& coords)
{

  assert (coords.size() == Protein::NUMDOF);
  assert ((coords[0] >= -Protein::maxX) && (coords[0] <=  Protein::maxX));
  assert ((coords[1] >= -Protein::maxY) && (coords[1] <=  Protein::maxY));

  nodeClass * newNode;
  
  x = coords[0];
  y = coords[1];
  
  computeParams();
  computeEnergy();
  
  assert (energy < maxEnergy);

  myTree   = NULL;
  myNode   = NULL;
  numLinks = 0;
  index                   = inputIndex;
  newNode                 = new nodeClass;
  newNode->myConfig       = this;
  setMyNode(newNode);
  testInvariant();
}


//----------------------------------------------------------------------------
//----------------I/O---------------------------------------------------------

/* only prints parameters, and to stdout. */
void Protein::printParameters() const
{
  cout << params[0] << " " << params[1] << endl;
}
  
void Protein::printParameters(FILE * fout) const {

  fprintf(fout, "%f %f \n", params[0], params[1]);

  testInvariant();    
}


void Protein::reportEnergy() {
  cout << " energy = " << energy << endl;
  testInvariant();
}


//----------------------------------------------------------------------------
//------other methods---------------------------------------------------------

/**
  sets params array of the protein, which determines the relative location 
  of the node on the landscape
*/
void Protein::computeParams()
{
  vector<float> coords(2);
  
  coords[0] = x; coords[1] = y;
  coords2Params(params, coords);
}

///computes energy of protein.
float Protein::computeEnergy() 
{
   
  vector<float> coords(2);

  coords[0] = x; coords[1] = y;
  energy = computeConfigEnergy(coords);
  

  return energy;
}

/** 
    returns true if configuration does not collide with itself, and is low than a probabilistically generated energy threshold.
*/
bool Protein::probabilisticAcceptConformation() const
{
  assert (maxEnergy > minEnergy);

  if (energy > Protein::maxEnergy) {
    return false;
  }
  else 
    {
      float randomNumber        = random()/MY_RAND_MAX;
      assert ((randomNumber >= 0) && (randomNumber <= 1.0));
      float acceptanceThreshold = (energy - Protein::minEnergy) / 
	(Protein::maxEnergy - Protein::minEnergy);
      if (acceptanceThreshold < randomNumber) 
	{
	  return true;
	}
      else 
	{ 
	  return false;
	}
    }
}

inline void Protein::testInvariant() const throw()
{
  assert(index >= -1);
  /* index can be -1 */
  assert(energy  <= Protein::maxEnergy); 

  assert(params.size() == 2);
  assert((params[0] >= 0.0) && (params[0] <= 1.0) && (params[1] >= 0.0) && (params[1] <= 1.0));
  
  float myQ0, myQ1;
  
  myQ0 = (x - (-maxX)) / (maxX - (-maxX));
  myQ1 = (y - (-maxY)) / (maxY - (-maxY));

  assert (fabs(myQ0 - params[0] ) < 1E-3);
  assert (fabs(myQ1 - params[1] ) < 1E-3);

  assert (fabs(x) <= Protein::maxX);
  assert (fabs(y) <= Protein::maxY);
}

/** 
   returns the L2 distance between 2 proteins
*/
float Protein::computeDistance(const Protein * p) const
{
  return computeConfigDistance(getCoords(), p->getCoords());
}

///returns L2 distance between 2 configs
float Protein::computeDistance(const vector<float>& B) const
{
  return computeConfigDistance(getCoords(), B);
}

void Protein::loadEnergyFunctionParameters(const char * filename)
{
  FILE * fin = fopen(filename, "r");
  if (fin == NULL)
    cout << "could not read " << filename << endl;
  else
    cout << "reading " << filename << endl;
    
  assert(fin != NULL);

  float alpha , tau , xo , yo;


  while (!feof(fin))
    {
      fscanf(fin, "%f %f %f %f\n", &alpha , &tau , &xo , &yo);
      //      cout << "alpha = " << alpha << endl;
      alphas.push_back(alpha);
      taus.push_back(tau);
      xos.push_back(xo);
      yos.push_back(yo);
    }
  fclose(fin);
}


//-------------------------------------------------------------------
//------helper functions----------------------------------------------



//----------------------------------------------------------------------
//--------accessors, mutators-------------------------------------------


void   Protein::setNumLinks    (int n)
{
  numLinks = n;
}

void   Protein::setIndex(int index)
{
  this->index = index;
}

float Protein::getX() const
{
  return x;
}

float Protein::getY() const
{
  return y;
}


int Protein::getIndex() const 
{
  //  testInvariant();
  return index;
}

/** 
   returns num. neighbors 
*/
int Protein::getNumLinks() const 
{
  testInvariant();
  //  return neighbours.size();
  return numLinks;
}
///
vector<float> Protein::getCoords() const
{
  vector<float> retval(2);
  retval[0] = x; retval[1] = y;
  return retval;
}



void Protein::setMyNode (nodeClass * node) {     // sets the myNode variable to the input parameter
  myNode = node;
  //  testInvariant(); commented out since protein constructor calls setMyNode
  // and the configs created in the protein constructor may have high energy
  // which makes testInvariant angry.
  
}

// returns true if ith amino acid is in loop structure. i = 0.. numAAs-1

void Protein::setEnergy(float iEnergy) {
    energy = iEnergy;
    testInvariant();
}




float Protein::getEnergy() const {
  return energy;
}

/** 
    only sets x and y.
*/
void   Protein::setXY (float x, float y)
{
  this->x = x;
  this->y = y;
}

vector<float> Protein::getParams() const
{
  return params;
}

///accessor.
nodeClass * Protein::getMyNode()
{
  return myNode;
}


//----------------------------------------------------------------------------
//-----------energy computation related---------------------------------------

/* 
   given x, y;
   this method computes the contribution of the energy well
   to the energy function.
   Note that it is not read from any file, so each time the parameters
   of the well around the folded state is changed, this function has to
   be changed as well.
*/
inline float  foldedStateTerm (float x,float y)
{
  float retval    = 0.0;
  const int L     = 19;// the peak around folded state
  const int K     = 50;// the pit at the folded state
  const float r_o = sqrt (K + L);// radius of the well
  const float tau = 10;// how quickly the energy barrier decays
  float s         = x * x + y * y; // sqr of the sum.
  float sqrt_s    = sqrt (s);
  if (sqrt_s < r_o)
    retval        = retval + s - K;
  else
    retval        = retval + L * exp ( - (sqrt_s - r_o) / tau);
  
  return retval;
}


/** returns energy. 
    also scales energy by 0.1 ( so that it lies between -5 .. 5)
*/
float computeConfigEnergy(const vector<float>& coords)
{
  float alpha , tau, xo, yo, x, y;

  float energy = 0.0;
  
  assert (coords.size() == Protein::NUMDOF);
  assert ((coords[0] <= Protein::maxX) && (coords[0] >= -Protein::maxX));
  assert ((coords[1] <= Protein::maxY) && (coords[1] >= -Protein::maxY));
  

  x = coords[0] ; y = coords[1];
  
  // the below is commented out since 
  // in the line minimization, the initial bracket may not be 
  // within the landscape bounds.
  //  assert(fabs(x) <= Protein::maxX); 
  //  assert(fabs(y) <= Protein::maxY);

  for (unsigned int i = 0; i < Protein::alphas.size(); i++)
    {
      alpha   = Protein::alphas[i];
      tau     = Protein::taus  [i];
      xo      = Protein::xos   [i];
      yo      = Protein::yos   [i];
      energy += energyTerm      (alpha,tau,xo,yo,x,y);
      //       cout << energy << endl;
    }
   
   energy     += foldedStateTerm (x,y);
   //     cout << energy << endl;
   Protein::numEnergyComputations ++;
   energy     *= 0.1;
   return energy;
}




/** xI and yI corresponds to the location of the node, x and y. */
inline float energyTerm (float alpha,float tau,  float x_o, float y_o, 
			 float xI, float yI)
{
  
  assert(tau != 0.0);

  float x      = (xI - x_o)/tau;
  float y      = (yI - y_o)/tau;
  
  
  return (alpha * exp(-x * x - y * y));
}



//--------node sampling----------------------------------------------

/** 
   generates a random config. around the original.
   the parameters are within +-Protein::deltaX,deltaY of the original.

   the final result is guaranteed to be within the bounds of the 
   landscape, but not necessarily collision free.
*/
void sampleAroundCoords(vector<float> & newCoords, 
			const vector<float> & coords)
{
  assert (coords.size() == 2);
  assert (Protein::deltaX == Protein::deltaY); 
  // not required, but assumed for the current implementation to work.
  assert (Protein::maxX   == Protein::maxY);     //ditto
  assert (fabs(coords[0]) <= Protein::maxX);
  assert (fabs(coords[1]) <= Protein::maxX);
  

  newCoords.resize(coords.size());
  for (unsigned int i = 0; i < coords.size(); i++)
    {
      float up  = coords[i] + Protein::deltaX;
      float lo  = coords[i] - Protein::deltaX;
      if ( up > Protein::maxX )  up = Protein::maxX;
      if ( lo < -Protein::maxX)  lo = -Protein::maxX;
      newCoords[i] = lo + (up-lo)*random()/MY_RAND_MAX;
    }
  
  assert ((newCoords[0] >= -Protein::maxX) && (newCoords[0] <= Protein::maxX));
  assert ((newCoords[1] >= -Protein::maxY) && (newCoords[1] <= Protein::maxY));

}


///returns coordinates within the boundaries of the landscape, but not necessarily collision-free.
void sampleRandomCoords(float * x, float * y)
{
  float rN = random() / MY_RAND_MAX;
  *x        = 2 * (rN - 0.5) * Protein::maxX;
  rN       = random() / MY_RAND_MAX;
  *y        = 2 * (rN - 0.5) * Protein::maxY;

}

/**
   returns a vector s.t. the energy is < maxEnergy.
*/
vector<float> generateRandomConfig ()
{
  vector<float> coords(2);
  int       iter    = 0;
  const int maxIter = 100;
  bool        found = false;

  while ((!found) && (iter < maxIter))
    {
      sampleRandomCoords(&coords[0],&coords[1]);
      float energy = computeConfigEnergy(coords);
      if (energy < Protein::maxEnergy)
	found = true;
      iter ++;
    }
  
  if (!found)
    {
      cout << "couldn't find a start config after " << iter << " iterations. exiting.." << endl;
      exit(0);
    }
  return coords;
}

//-----------------------------------------------------------------------
//-----other helper functions -------------------------------------------


///
void coords2Params      (vector<float> & newParams, 
			 const vector<float> & coords)
{
  assert (coords.size() == Protein::NUMDOF);
  
  newParams.resize(2);
  newParams[0] = (coords[0] - (-Protein::maxX)) / (Protein::maxX - (-Protein::maxX));
  newParams[1] = (coords[1] - (-Protein::maxY)) / (Protein::maxY - (-Protein::maxY));  

  assert ((newParams[0] >= 0) && (newParams[0] <= 1));
  assert ((newParams[1] >= 0) && (newParams[1] <= 1));
  
}

///
void params2Coords (vector<float>& coords, 
		    const vector<float> & sampled_params)
{
  assert (sampled_params.size() == Protein::NUMDOF);
  assert ((sampled_params[0] >= 0) && (sampled_params[0] <= 1));
  assert ((sampled_params[1] >= 0) && (sampled_params[1] <= 1));

  coords.resize(2);
  coords[0]        = 2 * (sampled_params[0] - 0.5) * Protein::maxX;
  coords[1]        = 2 * (sampled_params[1] - 0.5) * Protein::maxY;
}





/** 
    just average of the start and end.
    can be used both for coordinates and normalized parameters.
*/
void midConfig (vector<float> & mid, 
		const vector<float> & start,
		const vector<float> & end)
{
  assert (Protein::NUMDOF == 2);
  assert (start.size()    == 2);
  assert (end.size()      == 2);


  mid.resize(2);
  mid[0] = (start[0] + end[0]) / 2.0;
  mid[1] = (start[1] + end[1]) / 2.0;
}

float computeConfigDistance(const vector<float>& A, const vector<float>& B) 
{
  assert(B.size() == Protein::NUMDOF);
  assert(A.size() == Protein::NUMDOF);
  float deltaX = B[0] - A[0];
  float deltaY = B[1] - A[1];
  return (sqrt(deltaX * deltaX +   deltaY * deltaY));
}



///  returns true if the distance between configs < epsilon.
bool configsClose(const vector<float>&  config1, 
		  const vector<float>&  config2,
		  const float epsilon)
{
  assert (config1.size() == 2);
  assert (config2.size() == 2);
  float dist = computeConfigDistance(config1, config2);
  return (dist < epsilon);
}

/// returns true if the distance between configs < epsilon.
bool configsClose(const Protein * p1, 
		  const Protein * p2,
		  const float epsilon)
{
  return (configsClose(p1->getCoords(), p2->getCoords(), epsilon));
}


/// returns true if collision free.
bool checkCollision(float energy)
{
  if (energy > Protein::maxEnergy) 
    {
      return false;
    }
  else 
    {
      return true; 
    }
}

ostream& operator<< (ostream & output, const Protein& p) 
{
  output << "Printing protein..." << endl;
  output << "x = " << p.x << " y = " << p.y << endl;
  output << "energy = " << p.getEnergy() << endl;
  output << "params : " << p.params[0] << " " << p.params[1] << endl;
  return output;
}



//@}
